@@ -48,7 +48,6 @@
 
 #define KE_TIME_WAIT 			5000
 #define COUNT_OF_ITEM_IN_LAYER_KEYSET 	16
-#define COUNT_OF_ITEM_IN_KEYECHO_KEYSET 16
 
 #define MODIFIER(K,C) (K ? C : "")
 #define LINE(X)       (X != 0 ? "-" : "")
@@ -69,24 +68,12 @@ enum
 
 static gint 			ke_log_flags;
 
-#define KE_INTERESTED_MODIFIERS_NUMBER 6
-struct
+typedef struct
 {
-    gint  	modifier;
-    KeySym 	keysym;
-    KeyCode     keyCode;
-    u_int32_t   byteIndex;
-    u_int32_t   bitIndex;
-} KeModifierMap [KE_INTERESTED_MODIFIERS_NUMBER] =
-{
-    {SPI_KEYMASK_SHIFT, 	XK_Shift_L, 	0, 0, 0},
-    {SPI_KEYMASK_SHIFT, 	XK_Shift_R, 	0, 0, 0},
-    {SPI_KEYMASK_CONTROL, 	XK_Control_L, 	0, 0, 0},
-    {SPI_KEYMASK_CONTROL, 	XK_Control_R, 	0, 0, 0},
-    {SPI_KEYMASK_ALT, 		XK_Alt_L, 	0, 0, 0},
-    {SPI_KEYMASK_ALT, 		XK_Alt_R, 	0, 0, 0}
-};
-
+    gint    flag;
+    SREvent *evnt;
+} IdleStruct;    
+    
 /**
  *
  * Key string translation table.
@@ -234,6 +221,31 @@ typedef enum
 **/
 static KEStateEnum ke_keyboard_status = KE_IDLE;
 
+IdleStruct* ke_idle_struct_new ()
+{
+    IdleStruct* ke_idle;
+    ke_idle = g_new0 (IdleStruct, 1);
+    return ke_idle;
+}
+
+void ke_idle_struct_free (IdleStruct *ke_idle)
+{
+    sre_release_reference (ke_idle->evnt);
+    g_free (ke_idle);
+}
+
+static gboolean
+ke_repot_layer_cmd (gpointer data)
+{
+    IdleStruct *ke_idle = (IdleStruct*)data;
+
+    if (ke_keyboard_event_sink_cb != NULL) 
+	ke_keyboard_event_sink_cb (ke_idle->evnt, ke_idle->flag);
+	
+    ke_idle_struct_free (ke_idle);
+    return FALSE;
+}
+
 /**
  * ke_log_at_spi
  *
@@ -369,10 +381,12 @@ ke_return_keystring (gint keyID)
         case XK_Down:	rv = g_strdup("Down");break;
         case XK_Page_Up:rv = g_strdup("Page_Up");break;
         case XK_Page_Down:rv = g_strdup("Page_Down");break;
+	case XK_BackSpace:rv = g_strdup ("BackSpace");break;
+	case XK_Delete:rv = g_strdup ("Delete");break;
+	case XK_Escape:rv = g_strdup ("Escape");break;
         default:
     	    rv = NULL;
     }
-    
     sru_debug ("Return keystring:%s",rv);
     
     return rv;
@@ -398,6 +412,7 @@ ke_call_srkey_cb (gulong keyID,
 {
     SREvent *evnt = NULL;
     SRHotkeyData *srhotkey_data = NULL;
+    IdleStruct *ke_idle;
     evnt = sre_new ();
     sru_return_if_fail (evnt);
     srhotkey_data = (SRHotkeyData *) g_new0 (SRHotkeyData, 1);
@@ -424,9 +439,14 @@ ke_call_srkey_cb (gulong keyID,
 	ke_log_gnopernicus (evnt);
 	
 	if (ke_keyboard_event_sink_cb != NULL) 
-	    ke_keyboard_event_sink_cb (evnt,0);
-    }
-    
+	{
+	    sre_add_reference (evnt);
+	    ke_idle = ke_idle_struct_new ();
+	    ke_idle->flag = 0;
+	    ke_idle->evnt = evnt;
+	    g_idle_add (ke_repot_layer_cmd, ke_idle);
+	}
+    }	
     sre_release_reference (evnt);
 }
 
@@ -453,8 +473,8 @@ ke_report_keyecho_event (const AccessibleKeystroke *key, gpointer user_data)
 	     (guint) key->modifiers,
 	     (gint) key->keycode,
 	     (long int) key->timestamp,
-	     key->keystring);   
-
+	     key->keystring);  
+	     
     ke_log_at_spi (key);    
     
     if (key->type == SPI_KEY_RELEASED)
@@ -566,6 +586,7 @@ ke_return_key (glong keyID)
 }
 
 
+
 /**
  *
  * ke_call_keyboard_layer_cb
@@ -584,6 +605,7 @@ ke_call_keyboard_layer_cb (const gchar *buf,
 			   SREventType  event_type)
 {
     SREvent *evnt = NULL;
+    IdleStruct *ke_idle;
     
     sru_return_if_fail (buf);
     sru_return_if_fail (strlen(buf) != 0);
@@ -606,9 +628,14 @@ ke_call_keyboard_layer_cb (const gchar *buf,
 
 	/* call event sink from srcore */
         if (ke_keyboard_event_sink_cb != NULL) 
-	    ke_keyboard_event_sink_cb (evnt, flag);
-    }
-	
+	{
+	    sre_add_reference (evnt);
+	    ke_idle = ke_idle_struct_new ();
+	    ke_idle->flag = flag;
+	    ke_idle->evnt = evnt;
+	    g_idle_add (ke_repot_layer_cmd, ke_idle);
+	}
+    }	
     sre_release_reference (evnt);
 }
 
@@ -980,52 +1007,6 @@ ke_keysym_to_string (glong keysym, gint modifier)
     return keystring;
 }
 
-/**
- * ke_return_active_modifiers
- *
- *
- * Return state of Shifts, Alts, Controls and Num Lock.
- *
-**/
-gint
-ke_return_active_modifiers (void)
-{
-    gchar       keyVector[32];
-    Display    *display;
-    gint        i;
-    gint 	rv = 0;
-
-    if ((display = XOpenDisplay(NULL) ) == NULL ) 
-    {
-        sru_warning ("Unable to open display.");
-        return 0;
-    }
-    
-    /* Return the keyvector */
-    XQueryKeymap (display, keyVector);
-        
-    /*Return state of Shifts, Alts, Controls and for the Num Lock */
-    for (i = 0 ; i < G_N_ELEMENTS (KeModifierMap) ; i++)
-    {
-	KeModifierMap[i].keyCode = XKeysymToKeycode (display, KeModifierMap[i].keysym);
-
-	if (KeModifierMap[i].keyCode == 0) 
-	{
-    	    sru_warning ("Invalid KeyCode.");
-    	    return 0;
-	}
-                                                                                
-	KeModifierMap[i].byteIndex = (KeModifierMap[i].keyCode/8);
-	KeModifierMap[i].bitIndex  = KeModifierMap[i].keyCode & 7;
-	
-	if (1 & (keyVector[KeModifierMap[i].byteIndex] >> KeModifierMap[i].bitIndex)) 
-	    rv = rv | KeModifierMap[i].modifier;
-
-    }
-    
-    return rv;
-}
-
 /**
 *
 * ke_report_user_key  
@@ -1043,7 +1024,6 @@ ke_report_user_key_event (const AccessibleKeystroke *key, gpointer user_data)
     static gboolean user_busy = FALSE;    
     static GQueue  *user_queue = NULL;
     KEKeyEvent     *ev;
-
     sru_debug ("(User)Received key event:\n\tsym %ld (%c)\n\tmods %x\n\tcode %d\n\ttime %ld\n\tkeystring \"%s\"\n\ttype %d (press = %d, release = %d)",
 	      (glong) key->keyID, (gchar)key->keyID,
 	      (guint) key->modifiers,
@@ -1095,7 +1075,10 @@ ke_report_user_key_event (const AccessibleKeystroke *key, gpointer user_data)
 	/* pop from queue the event */
 	key1 = (KEKeyEvent *) g_queue_pop_tail (user_queue);
 	/* get the status of modifiers */
-	modif  = ke_return_active_modifiers ();
+	if (key->modifiers & (~(SPI_KEYMASK_ALT | SPI_KEYMASK_CONTROL | SPI_KEYMASK_SHIFT)))
+	    sru_warning ("Callback invoked with an unexpected modifier.");
+	modif  = key1->modifiers & (SPI_KEYMASK_ALT | SPI_KEYMASK_CONTROL | SPI_KEYMASK_SHIFT);
+	
 	/* get the keystring */
 	keystr = ke_keysym_to_string (key1->keyID, modif);
         /* create a command */
@@ -1198,24 +1181,6 @@ ke_register_keysetitem (KeySetItem *ksi)
 				    			(gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
 				    			SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
     ke_print_register_return_value (retval, "");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_user_key_listener,
-				    			(AccessibleKeySet *) ksi->keyset,
-				    			ksi->modifier | SPI_KEYMASK_BUTTON1,
-				    			(gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				    			SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_user_key_listener,
-				    			(AccessibleKeySet *) ksi->keyset,
-				    			ksi->modifier | SPI_KEYMASK_BUTTON2,
-				    			(gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				    			SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_user_key_listener,
-				    			(AccessibleKeySet *) ksi->keyset,
-				    			ksi->modifier | SPI_KEYMASK_BUTTON3,
-				    			(gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				    			SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "");
 
     return TRUE;
 }
@@ -1305,60 +1270,6 @@ ke_layer_register_events (AccessibleKeystrokeListener *ke_layer_listener,
 	Layer key listener. Listen key event from numpad. Listen only if a numlock 
 	is active.
     */
-#ifndef USE_ALL_WINDOWS
-    retval = SPI_registerAccessibleKeystrokeListener (ke_layer_listener,
-				      (AccessibleKeySet *) ke_layer_keyset,
-				      SPI_KEYMASK_BUTTON1 | SPI_KEYMASK_NUMLOCK,
-				      (gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "(B1|NUMLOCK)");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_layer_listener,
-				      (AccessibleKeySet *) ke_layer_keyset,
-				      SPI_KEYMASK_BUTTON2 | SPI_KEYMASK_NUMLOCK,
-				      (gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "(B2|NUMLOCK)");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_layer_listener,
-				      (AccessibleKeySet *) ke_layer_keyset,
-				      SPI_KEYMASK_BUTTON3 | SPI_KEYMASK_NUMLOCK,
-				      (gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-			    	      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "(B3|NUMLOCK)");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_layer_listener,
-				      (AccessibleKeySet *) ke_layer_keyset,
-				      SPI_KEYMASK_BUTTON1 | 
-				      SPI_KEYMASK_BUTTON2 |
-				      SPI_KEYMASK_NUMLOCK,
-				      (gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "(B1|B2|NUMLOCK)");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_layer_listener,
-				      (AccessibleKeySet *) ke_layer_keyset,
-				      SPI_KEYMASK_BUTTON1 | 
-				      SPI_KEYMASK_BUTTON3 |
-				      SPI_KEYMASK_NUMLOCK,
-				      (gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "(B1|B3|NUMLOCK)");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_layer_listener,
-				      (AccessibleKeySet *) ke_layer_keyset,
-				      SPI_KEYMASK_BUTTON3 |
-				      SPI_KEYMASK_BUTTON2 |
-				      SPI_KEYMASK_NUMLOCK,
-				      (gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "(B2|B3|NUMLOCK)");
-    retval = SPI_registerAccessibleKeystrokeListener (ke_layer_listener,
-				      (AccessibleKeySet *) ke_layer_keyset,
-				      SPI_KEYMASK_BUTTON3 |
-				      SPI_KEYMASK_BUTTON2 |
-				      SPI_KEYMASK_BUTTON1 |
-				      SPI_KEYMASK_NUMLOCK,
-				      (gulong) ( SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    ke_print_register_return_value (retval, "(B1|B2|B3|NUMLOCK)");
-#endif
-
     retval = SPI_registerAccessibleKeystrokeListener (ke_layer_listener,
 				      (AccessibleKeySet *) ke_layer_keyset,
 				      SPI_KEYMASK_NUMLOCK,
@@ -1455,22 +1366,6 @@ ke_keyecho_register_events (AccessibleKeystrokeListener *ke_key_listener,
 	SHIFT_L, SHIFT_R, Control_L, Control_R, Up, Down, Left, Right, 
 	PageUp, PageDown, Home, End, CapsLock, NumLock,
     */
-    retval = SPI_registerAccessibleKeystrokeListener (ke_key_listener,
-				      (AccessibleKeySet *) ke_key_keyset,
-				      SPI_KEYMASK_BUTTON1,
-				      (gulong) (SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    retval = SPI_registerAccessibleKeystrokeListener (ke_key_listener,
-				      (AccessibleKeySet *) ke_key_keyset,
-				      SPI_KEYMASK_BUTTON2,
-				      (gulong) (SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-    retval = SPI_registerAccessibleKeystrokeListener (ke_key_listener,
-				      (AccessibleKeySet *) ke_key_keyset,
-				      SPI_KEYMASK_BUTTON3,
-				      (gulong) (SPI_KEY_PRESSED | SPI_KEY_RELEASED ),
-				      SPI_KEYLISTENER_NOSYNC | SPI_KEYLISTENER_CANCONSUME);
-				      
     retval = SPI_registerAccessibleKeystrokeListener (ke_key_listener,
 				      (AccessibleKeySet *) ke_key_keyset,
 				      SPI_KEYMASK_UNMODIFIED,
@@ -1545,7 +1440,7 @@ ke_get_keypad_keycodes (void)
  * return: keysym array
  *
 **/
-long keyecho_keysym[COUNT_OF_ITEM_IN_KEYECHO_KEYSET] = { 
+long keyecho_keysym[] = { 
     XK_Alt_L, 
     XK_Alt_R,
     XK_Shift_L,
@@ -1561,7 +1456,10 @@ long keyecho_keysym[COUNT_OF_ITEM_IN_KEYECHO_KEYSET] = {
     XK_Up,
     XK_Down,
     XK_Page_Up,
-    XK_Page_Down
+    XK_Page_Down,
+    XK_BackSpace,
+    XK_Delete,
+    XK_Escape
 };
 
 gchar*
@@ -1569,10 +1467,14 @@ ke_get_keyecho_keysyms (void)
 {
     gchar *rv = NULL;
     gint i;
+    Display *disp;
+    disp = GDK_DISPLAY ();
 
     for (i = 0 ; i < G_N_ELEMENTS(keyecho_keysym) ; i++)
     {
-        if (g_unichar_validate ((guint32)keyecho_keysym[i]))
+	short key;
+	key = XKeysymToKeycode (disp, keyecho_keysym[i]);
+	if (g_unichar_validate ((guint32)key)) 
         {
 	    gchar *keysym = NULL;
     	    gchar *tmp = NULL;
@@ -1728,7 +1630,7 @@ ke_init (KeyboardEchoCB kecb)
 	Keyboard echo.
     */
     ke_keyecho_keysyms = ke_get_keyecho_keysyms ();
-    ke_keyecho_keyset  = SPI_createAccessibleKeySet (COUNT_OF_ITEM_IN_KEYECHO_KEYSET, ke_keyecho_keysyms, NULL, NULL);
+    ke_keyecho_keyset  = SPI_createAccessibleKeySet (G_N_ELEMENTS (keyecho_keysym),ke_keyecho_keysyms, NULL, NULL);
     g_free (ke_keyecho_keysyms);
 
     /*
@@ -1779,18 +1681,6 @@ ke_user_key_list_unregister_events (void)
 	SPI_deregisterAccessibleKeystrokeListener (ke_user_key_listener, 
 						   ksi->modifier);
 
-	SPI_deregisterAccessibleKeystrokeListener (ke_user_key_listener, 
-						   ksi->modifier | 
-						   SPI_KEYMASK_BUTTON1);
-
-	SPI_deregisterAccessibleKeystrokeListener (ke_user_key_listener, 
-						   ksi->modifier |
-						   SPI_KEYMASK_BUTTON2);
-
-	SPI_deregisterAccessibleKeystrokeListener (ke_user_key_listener, 
-						   ksi->modifier |
-						   SPI_KEYMASK_BUTTON3);
-	
 	ke_key_set_item_free (ksi);
 	
 	elem->data = NULL;
@@ -1808,35 +1698,6 @@ ke_user_key_list_unregister_events (void)
 static void 
 ke_layer_unregister_events (void)
 {
-#ifndef USE_ALL_WINDOWS
-    SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
-					    SPI_KEYMASK_BUTTON1 |
-					    SPI_KEYMASK_NUMLOCK);
-    SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
-					    SPI_KEYMASK_BUTTON2 |
-					    SPI_KEYMASK_NUMLOCK);
-    SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
-					    SPI_KEYMASK_BUTTON3 |
-					    SPI_KEYMASK_NUMLOCK);
-    SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
-					    SPI_KEYMASK_BUTTON1 |
-					    SPI_KEYMASK_BUTTON2 |
-					    SPI_KEYMASK_NUMLOCK);
-    SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
-					    SPI_KEYMASK_BUTTON1 |
-					    SPI_KEYMASK_BUTTON3 |
-					    SPI_KEYMASK_NUMLOCK);
-    SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
-					    SPI_KEYMASK_BUTTON2 |
-					    SPI_KEYMASK_BUTTON3 |
-					    SPI_KEYMASK_NUMLOCK);
-    SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
-					    SPI_KEYMASK_BUTTON1 |
-					    SPI_KEYMASK_BUTTON2 |
-					    SPI_KEYMASK_BUTTON3 |
-					    SPI_KEYMASK_NUMLOCK);
-#endif
-
     SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
 					    SPI_KEYMASK_NUMLOCK);
     SPI_deregisterAccessibleKeystrokeListener (ke_layer_listener, 
@@ -1869,12 +1730,6 @@ static void
 ke_keyecho_unregister_events (void)
 {
     /* deregister keylisteners */
-    SPI_deregisterAccessibleKeystrokeListener (ke_keyecho_listener, 
-					    SPI_KEYMASK_BUTTON1);
-    SPI_deregisterAccessibleKeystrokeListener (ke_keyecho_listener, 
-					    SPI_KEYMASK_BUTTON2);
-    SPI_deregisterAccessibleKeystrokeListener (ke_keyecho_listener, 
-					    SPI_KEYMASK_BUTTON3);
     SPI_deregisterAccessibleKeystrokeListener (ke_keyecho_listener, 
 					    SPI_KEYMASK_UNMODIFIED );
     SPI_deregisterAccessibleKeystrokeListener (ke_keyecho_listener, 
